/*
** [DOC File]
**
** [Name]				grpstrip.c
** [Description]     GRP Stripper
** [Author]				Rex Engel
** [Version]			1.0
** [Start Date]		01 May 1996
**
**	[Revision History]
**
** Rev	Who		When			What
** ==========================================================================
** 1.0	Rex      01-05-1996	Initial Revision
**
** [DOC End]
**
*/

/*
** ---------------------------------------------------------------------------
** Includes
** ---------------------------------------------------------------------------
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <memory.h>
#include <dos.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <share.h>
#include <io.h>


/*
** ---------------------------------------------------------------------------
** Defines
** ---------------------------------------------------------------------------
*/
#define BUFSIZE	32767
#define TRUE		1
#define FALSE		0
#define strncopy( t, s, n )	memcpy( (t), (s), (n) ); t[ n ] = '\0';

/*
** ---------------------------------------------------------------------------
** Typedefs
** ---------------------------------------------------------------------------
*/
#pragma -a-
typedef struct {
	char 			  name[ 12 ];
	unsigned long length;
} ENTRY;
#pragma -a

typedef enum {
	UNKNOWN = 0,
	EXTRACT = 1,
	LIST = 2
} MODE;

/*
** ---------------------------------------------------------------------------
** Functions
** ---------------------------------------------------------------------------
*/
static int match( char *name, char *patt )
{
	while (*patt != '\0') {								// empty pattern => end
		if (( *patt == '*' ) &&
			 ( match( name, patt + 1))) {				// RECURSE: try empty string
			return TRUE;									// EXIT: it's a hit...
		}
		if (*name == '\0') {								// empty dish
			return FALSE;									// EXIT: end of name
		}
		switch (*patt) {
			case '*':										// -> wild string
				break;										// ignore char
			case '?':										// -> wild char
				patt++;										// char matched
				break;
			default:											// -> normal char
				if (toupper(*patt) != toupper(*name)) { // compare char
					return FALSE;							// EXIT: char mismatch
				}
				patt++;										// char matched
				break;
		}
		name++;												// step forward
	}
	return (*name == '\0');
}

static void extract( int sfh, unsigned long offset, char *filename, unsigned long length )
{
	int tfh;
	unsigned long fptr, save;
	char *buffer;

	if ( ( buffer = malloc( BUFSIZE ) ) == NULL ) {
		fprintf( stderr, "Cannot allocate copy buffer\n" );
		return;
	}
	if ( ( tfh = sopen( filename, O_WRONLY | O_BINARY | O_CREAT, S_IWRITE, SH_DENYNO ) ) < 0 ) {
		fprintf( stderr, "Cannot create file: %s\n", filename );
		free( buffer );
		return;
	}
	save = tell( sfh );
	lseek( sfh, offset, SEEK_SET );
	for ( fptr = 0; fptr < length; fptr += min( length - fptr, BUFSIZE ) ) {
		read ( sfh, buffer, (int)min( length - fptr, BUFSIZE ) );
		write( tfh, buffer, (int)min( length - fptr, BUFSIZE ) );
	}
	printf( "Extracted %s\n", strlwr( filename ) );
	close( tfh );
	free( buffer );
	lseek( sfh, save, SEEK_SET );
	chmod( filename, S_IWRITE );
}

int main( int argc, char *argv[] )
{
	ENTRY header, entry;
	int fh, i;
	char filename[ FILENAME_MAX ] = "duke3d.grp";
	char foundname[ 13 ];
	unsigned long offset;
	MODE mode = UNKNOWN;

	fprintf( stderr, "GRPSTRIP 1.0 By Rex Engel ( rex@worldaccess.nl )\n\n" );

	switch ( toupper( *( argv[ 1 ] ) ) ) {
		case 'X' :
			mode = EXTRACT;
			break;
		case 'L' :
			mode = LIST;
			break;
		case 'H' :
 			printf( "Usage: grpstrip L|X|H <pattern>\n\n" );
			printf( "Commands:\n" );
			printf( "X Extract files\n" );
			printf( "L Shows list\n"    );
			printf( "H Shows this help\n\n" );
			printf( "Pattern is the pattern to match\n" );
			exit( 0 );
	}

 	if ( argc != 3 ) {
 		fprintf( stderr, "Usage: grpstrip L|X|H <pattern>\n" );
		fprintf( stderr, "\n" );
		fprintf( stderr, "Example : grpstrip L *.mid\n" );
 		exit( 0 );
 	}
	if ( mode == UNKNOWN ) {
 		fprintf( stderr, "Unknown command\n" );
 		exit( 0 );
	}

	if ( ( fh = sopen( filename, O_RDONLY | O_BINARY, S_IREAD ) ) < 0 ) {
		fprintf( stderr, "Cannot open grp file\n" );
		exit( 0 );
	}
	lseek( fh, 0L, SEEK_SET );
	read( fh, &header, sizeof( ENTRY ) );

	printf( "File %s contains %d entries\n", filename, header.length );

	offset = ( header.length + 1 ) * sizeof( ENTRY ) ;
	for ( i = 0; i < header.length ; i++ ) {

		read( fh, &entry, sizeof( ENTRY ) );

		strncopy( foundname, entry.name, 12 );

		if ( match( foundname, argv[ 2 ] ) ) {
			switch ( mode ) {
				case EXTRACT :
					extract( fh, offset, foundname, entry.length );
					break;
				case LIST :
					printf( "%-12s %7ld bytes\n", foundname, entry.length );
					break;
			}
		}
		offset += entry.length;
	}

	close( fh );
	return 0;
}
